#!/usr/bin/env python3

import numpy as np
import ezdxf 
import random
from math import sqrt
import copy
import warnings
unit_convert=10000 #ses to dxf

class pre_process():
    def __init__(self,dxf_file):
        """
        Runneg this class will generate a temporary file that is ready to be sent to auto-router
        Please make sure the ME drawing has put everything need to cut (obstacles) in "Cut" layer
        """
        self.dwg=ezdxf.readfile(dxf_file)
        self.msp=self.dwg.modelspace()
        self.savename='dwg_for_autorouter.dxf'

        self.create_layer('Circuit_Cut',6)
        self.create_layer('Circuit_Etc',8)

        self.keep_cut_layer_only()
        self.line_to_path()
        self.bry_ind=self.find_bry_index()

    def create_layer(self,layer_name,color):
        if not layer_name in self.dwg.layers:
            self.dwg.layers.new(name=layer_name,dxfattribs={'color':color})

    def keep_cut_layer_only(self):
        for e in self.msp.query('*[layer!="Cut"]'):
            self.msp.delete_entity(e)
    def line_to_path(self):
        # Write file init
        self.dwgWrite = ezdxf.new('R2010')  # create a new DXF R2010 drawing, official DXF version name: 'AC1024'
        mspWrite =self.dwgWrite.modelspace()  # add new entities to the model space

        # Variable init
        width = 0.26
        xAbMax = 0 
        yAbMax = 0 
        xAbMin = 1000
        yAbMin = 1000

        # Layer creation
        for layer in self.dwg.layers:
            name = layer.dxf.name
            lineType = layer.dxf.linetype
            color = layer.get_color()

            self.dwgWrite.layers.new(name=name, dxfattribs={'linetype': lineType, 'color': color})

        # loop for path creation
        for e in self.msp.query('LINE'):
            
            # finding info from read file
            x1 = e.dxf.start[0]
            x2 = e.dxf.end[0]
            y1 = e.dxf.start[1]
            y2 = e.dxf.end[1]
            layerName = e.dxf.layer

            # Calculations 
            yDiff = y2-y1
            yAbsDiff = abs(yDiff)
            yDiffDiv100 = yAbsDiff/100

            xDiff = x2-x1
            xAbsDiff = abs(xDiff)
            xDiffDiv100 = xAbsDiff/100

            yMin = min(y1, y2)
            xMin = min(x1, x2)

            yMax = max(y1, y2)
            xMax = max(x1, x2)

            # Calculations for bounding box
            if(yMin < yAbMin):
                yAbMin = yMin

            if(xMin < xAbMin):
                xAbMin = xMin

            if(yMax > yAbMax):
                yAbMax = yMax

            if(xMax > xAbMax):
                xAbMax = xMax

            # Line Cases 
            if (x1 == x2):
                writePoints = [(x1, yMin), (x1, yMax), ((x1+width),yMax), ((x1+width), yMin), (x1, yMin)]
            elif (y1 == y2):
                writePoints = [(xMin, y1), (xMin, (y1 + width)), (xMax,(y1 + width)), (xMax, y1), (xMin, y1)]

            elif ((x1 < x2 and y1 < y2 ) or (x1 > x2 and y1 > y2)):
                writePoints = [(xMin, yMin),((xMin - yDiffDiv100), (yMin + xDiffDiv100)), ((xMax - yDiffDiv100),(yMax + xDiffDiv100)), ((xMax), (yMax)), (xMin, yMin)]

            elif ((x1 < x2 and y1 > y2) or (x1 > x2 and y1 < y2)):
                writePoints = [(xMax, yMin),(xMin, yMax), ((xMin + yDiffDiv100),(yMax + xDiffDiv100)), ((xMax + yDiffDiv100), (yMin + xDiffDiv100)), (xMax, yMin)]

            # Write the line
            mspWrite.add_lwpolyline(writePoints, dxfattribs={'layer': layerName})


        xAbMax = xAbMax + 50
        yAbMax = yAbMax + 50
        xAbMin = xAbMin - 50
        yAbMin = yAbMin - 50
        self.boxPoints = [(xAbMin, yAbMin), (xAbMin, yAbMax), (xAbMax,yAbMax), (xAbMax, yAbMin), (xAbMin, yAbMin)]
        mspWrite.add_lwpolyline(self.boxPoints, dxfattribs={'layer': layerName})
        
        self.dwgWrite.saveas(self.savename)
    def find_bry_index(self):
        msp=self.dwgWrite.modelspace()
        bry=list(np.array(self.boxPoints).flatten())

        pts_list=[]
        for e in msp.query('LWPOLYLINE'):
            pts_list.append(np.array(e.get_points()))

        for i in range(len(pts_list)):
            pts_list[i]=pts_list[i][:,:2]
            pts_list[i]=pts_list[i].flatten()
            pts_list[i]=list(pts_list[i])
        
        ind=pts_list.index(bry)
        return ind


class post_process():
    def __init__(self,dxf_file,wiring_path,pin_at,trace_width,iso_size,cross_size):
        print('wiring path loaded, generating final drawing...')
        if not dxf_file.endswith('.dxf'):
            dxf_file+='.dxf'
        self.dxf_file=dxf_file
        self._wiring=wiring_path
        # self.pin_at=pin_at
        self.wiring=[]
        self.dwg=ezdxf.readfile(dxf_file)
        self.msp=self.dwg.modelspace()
        
        self.trace_width=trace_width #mm
        self.iso_size=iso_size #mm
        self.cross_cut_size=cross_size#mm
        self.trace_cut_width=0.25 #mm
        self.convert_unitfrom_ses()

        
        ## flatten pin_at_array
        self.pin_at=[]
        for module in pin_at:
            for pin in module:
                self.pin_at.append(pin)
        
        self.remove_redun_wire(0.4)
        self.modify_wire_and_merge()
        
        if not self.cross_cut_size==0:
            self.draw_cross_cut(self.cross_cut_size)
        if trace_width==0:
            self.draw_wires()
            if self.iso_size!=0:
                self.draw_rest_iso(self.iso_size,self.trace_cut_width)
        else:
            if self.iso_size==0:
                raise Exception ('Isolation box size is required as long as trace width is not equal to zero')
            else:    
                self.draw_complete_traces()

                self.draw_rest_iso(self.iso_size,self.trace_cut_width)

        self.dwg.saveas('dwg_w_circuit.dxf')
        print('drawing is ready for you to make a robot...')
    
    def draw_complete_traces(self):
        self.check_wiring_branch(self.trace_width)
        outer_helper=self.parallel_trace(self.trace_width,self.iso_size,'Circuit_Cut')
        
        trace_width=self.trace_width-self.trace_cut_width
        iso_size=self.iso_size-self.trace_cut_width
        
        self.check_wiring_branch(trace_width)
        self.parallel_trace(trace_width,iso_size,'Circuit_Etc')
        trace_width-=self.trace_cut_width
        iso_size-=self.trace_cut_width
        
        self.check_wiring_branch(trace_width)
        inner_helper=self.parallel_trace(trace_width,iso_size,'Circuit_Cut')
        for i in range(len(outer_helper)):
            self.msp.add_line(outer_helper[i],inner_helper[i],dxfattribs={
                'layer':'Circuit_Cut',
                'linetype':'DASHDOT'})
    
    def convert_unitfrom_ses(self):
        for path in self._wiring:
            p=np.array(path)/unit_convert
            # p[:,1]*=-1
            self.wiring.append(p.tolist())       
    
    def remove_redun_wire(self,tolerance):
        wiring=copy.deepcopy(self.wiring)
        self.wiring=[]
        for path in wiring:
            path_cp=copy.deepcopy(path)
            for i in range(len(path_cp)-1):
                pt1=path_cp[i]
                pt2=path_cp[i+1]

                if i<len(path_cp)-2:
                    pt3=path_cp[i+2]
                    if three_ptw_on_one_line(pt1,pt2,pt3):
                        path.remove(pt2)

                if distance_between_pts(pt1,pt2)<=tolerance and (not if_pts_are_vert_or_horz(pt1,pt2)) and pt1 in path and pt2 in path :
                        path.remove(pt2)
            if len(path)>=2:
                self.wiring.append(path)
    def modify_wire_and_merge(self):
        for i in range(len(self.wiring)):
            path=self.wiring[i]
            endpoints=[path[0],path[-1]]
            for j in range(len(endpoints)):
                ept=endpoints[j]
                if j==1:
                    k=-1
                    kk=-2
                else:
                    k=0
                    kk=1
                if len(path)>1:
                    further_ept=path[kk]              
                    for pin in self.pin_at:
                        if point_within_pinbox(ept,pin,self.iso_size):
                            self.wiring[i][k]=pin
                            if point_within_pinbox(further_ept,pin,self.iso_size):
                                self.wiring[i].pop(kk)
                            break
               
        path_list=copy.deepcopy(self.wiring)
        self.wiring=[]
        while len(path_list)!=0:
            merged_path=path_list[0]
            path_list.remove(merged_path)
            flag=True
            while flag:
                s=merged_path[0]
                e=merged_path[-1]
                flag=False
                for restpath in path_list:
                    s_=restpath[0]
                    e_=restpath[-1]
                    to_merge_copy=copy.deepcopy(restpath)
                    #update merge path
                    if s==s_:
                        merged_path.remove(s)
                        merged_path=list(reversed(merged_path))+(to_merge_copy)
                        path_list.remove(restpath)                
                        flag=True
                        break
                    if s==e_:
                        merged_path.remove(s)
                        merged_path=to_merge_copy+(merged_path)
                        path_list.remove(restpath)                
                        flag=True
                        break
                    if e==s_:
                        merged_path.remove(e)
                        merged_path=(merged_path)+(to_merge_copy)
                        path_list.remove(restpath)                
                        flag=True
                        break
                    if e==e_:
                        merged_path.remove(e)
                        merged_path=to_merge_copy+(list(reversed(merged_path)))
                        path_list.remove(restpath)                
                        flag=True
                        break
                    
            self.wiring.append(merged_path)

                    
    def draw_wires(self):
        for path in self.wiring:
            for i in range(len(path)-1):
                self.msp.add_line(path[i],path[i+1],dxfattribs={
                    'layer':'Circuit_Etc',
                    'linetype':'DASHDOT'})

    def check_wiring_branch(self,width):
        
        self.branch=[]
        self.branch_at_interval=[]
        self.branch_conn_at=[]
        self.path_at_pin=[]
        self.connected_pins=[]

        
        for i in range(len(self.wiring)):
            path=self.wiring[i]
            endpoints=[path[0],path[-1]]
            for j in range(len(endpoints)):
                ept=endpoints[j]
                connect_flag=False

                if j==1:
                        k=-1
                        contact_ept=path[-1]
                        further_ept=path[-2]
                elif j==0:
                    k=0
                    contact_ept=path[0]
                    further_ept=path[1]
                
                for pin in self.pin_at:
                    if distance_between_pts(ept,pin)<=(self.iso_size/2)*sqrt(2):
                        self.connected_pins.append(pin)
                        self.path_at_pin.append([i,k])
                        connect_flag=True
                        break
                if not connect_flag:
                    self.branch.append([i,k])
                    #look for the interval that connectes to such path endpoint
                    interval_flag=False
                    for r in range(len(self.wiring)):
                        restpath=self.wiring[r]
                        if restpath!=path:
                            for m in range(len(restpath)-1):
                                pt1=restpath[m]
                                pt2=restpath[m+1]
                                l1=find_line_eq(pt1,pt2)
                                if point_is_in_two_pts(ept,pt1,pt2) and if_pts_on_line(l1,ept):
                                    
                                    offset_lines=find_offset(l1,width)
                                    l2=find_line_eq(contact_ept,further_ept)
                                    dis=[]
                                    for off_line in offset_lines:
                                        inter=find_intersect(l2,off_line)
                                        if type(inter)!=type(None):
                                            dis_=distance_between_pts(further_ept,inter)
                                            dis.append(dis_)
                                    off_l=offset_lines[np.argmin(dis)]
                                    off_l2in,off_l2out=find_offset(l2,width)
                                    s_=find_intersect(off_l2in,off_l)
                                    e_=find_intersect(off_l2out,off_l)
                                    if distance_between_pts(pt1,s_)>distance_between_pts(pt1,e_):
                                        s=e_
                                        e=s_
                                    else:
                                        s=s_
                                        e=e_
                                    interval_flag=True
                                    break
                            if interval_flag:
                                break
                    if interval_flag:
                        self.branch_at_interval.append([r,m])
                        self.branch_conn_at.append([s,e])
                    else:
                        raise Exception ('path',path,k,'was detected as branch, but no connection with another path found')
            
        for path in self.wiring:
            for h in range(1,len(path)-2):
                pt1=path[h]
                pt2=path[h+1]
                for pin in self.pin_at:
                    if dis_between_pt_interval(pin,pt1,pt2)<(self.iso_size/2)*sqrt(2):
                        self.connected_pins.append(pin)
        
    def parallel_trace(self,width,iso_size,draw_layer):
        wiring=copy.deepcopy(self.wiring)
        helper_cut_list=[]
        for path in wiring:
            path_index=wiring.index(path)
            path_cp=copy.deepcopy(path)
            constrain_s,constrain_e=True,True
            
            if len(path)==1:
                continue
            ######        
            for i in range(len(path)-1):
                pt1=path[i][0],path[i][1]
                pt2=path[i+1][0],path[i+1][1]
                point_index=path_cp.index([pt1[0],pt1[1]])
                
                l1=find_line_eq(pt1,pt2)
                
                if i==0: #initalize
                        offl1_in,offl1_out=find_offset(l1,width)
                        
                        if [path_index,0] in self.branch:
                            branch_index=self.branch.index([path_index,0])
                            [s1,s2]=self.branch_conn_at[branch_index]
                        elif [path_index,0] in self.path_at_pin:
                            connect_index=self.path_at_pin.index([path_index,0])
                            pin=self.connected_pins[connect_index]
                            s1,s2,helper_pt=self.find_iso_bry(pt1,pt2,pin,iso_size,width,draw_layer,constrain_s)
                            helper_cut_list.append(helper_pt)
                        else:
                            raise Exception ('path start is neither a branch or a path connected to pin')

                if i<len(path)-2: #before last line
                    pt3=path[i+2][0],path[i+2][1]
                    l2=find_line_eq(pt2,pt3)
                #filter bi-sector
                    _bi=find_bi_sec_by_pts(pt1,pt2,pt3)
                    if len(_bi)!=1:
                        for sector in _bi:
                            if if_two_pt_on_diff_side(sector,pt1,pt3):
                                bi=sector
                    else: bi=_bi[0]

                    e1=find_intersect(offl1_in,bi)
                    e2=find_intersect(offl1_out,bi)
                else:                                    
                    if [path_index,-1] in self.branch:
                        branch_index=self.branch.index([path_index,-1])
                        [e1,e2]=self.branch_conn_at[branch_index]

                    elif [path_index,-1] in self.path_at_pin:
                        connect_index=self.path_at_pin.index([path_index,-1])
                        pin=self.connected_pins[connect_index]
                        e1,e2,helper_pt=self.find_iso_bry(pt2,pt1,pin,iso_size,width,draw_layer,constrain_e)
                        helper_cut_list.append(helper_pt)
                    else:
                        raise Exception ('path end is neither a branch or a path connected to pin')
                
                if if_two_pt_on_diff_side(l1,s1,e1):
                    e1_cp=e2
                    e2_cp=e1
                else:
                    e1_cp=e1
                    e2_cp=e2
                
                if [path_index,point_index] in self.branch_at_interval:
                    branch_index=self.branch_at_interval.index([path_index,point_index])
                    [s_sub,e_sub]=self.branch_conn_at[branch_index]

                    if not if_two_pt_on_diff_side(l1,s1,s_sub):

                        self.msp.add_line(s1,s_sub,dxfattribs={
                            'layer':draw_layer,
                            'linetype':'DASHDOT'})
                        self.msp.add_line(e_sub,e1_cp,dxfattribs={
                            'layer':draw_layer,
                            'linetype':'DASHDOT'})
                        self.msp.add_line(s2,e2_cp,dxfattribs={
                            'layer':draw_layer,
                            'linetype':'DASHDOT'})
                    else:

                        self.msp.add_line(s2,s_sub,dxfattribs={
                            'layer':draw_layer,
                            'linetype':'DASHDOT'})
                        self.msp.add_line(e_sub,e2_cp,dxfattribs={
                            'layer':draw_layer,
                            'linetype':'DASHDOT'})
                        self.msp.add_line(s1,e1_cp,dxfattribs={
                            'layer':draw_layer,
                            'linetype':'DASHDOT'})
                else:
                    self.msp.add_line(s1,e1_cp,dxfattribs={
                        'layer':draw_layer,
                        'linetype':'DASHDOT'})
                    self.msp.add_line(s2,e2_cp,dxfattribs={
                        'layer':draw_layer,
                        'linetype':'DASHDOT'})                 
                if i<len(path)-2:
                    offl1_in,offl1_out=find_offset(l2,width)
                    s1=find_intersect(offl1_in,bi)
                    s2=find_intersect(offl1_out,bi)
        return helper_cut_list
    def draw_cross_cut(self,size):
        for pin in self.pin_at:
            cross_cut(self.msp,size,pin)
    def find_iso_bry(self,contact_ept,further_ept,pin,size,width,draw_layer,constrain=True):

        l=find_line_eq(contact_ept,further_ept)
        parallel_traces=find_offset(l,width)
        four_l_points=isolation_box_linepts(size,pin)
        iso_points=isolation_box_linepts(size,pin,False)
        #diagnal eatching helper cut
        dis=[]
        for point in iso_points:
            dis_=distance_between_pts(further_ept,point)
            dis.append(dis_)
        diag_helper_cut_pt=iso_points[np.argmax(dis)]
        ######
        contact_list=[]
        for pl in parallel_traces:
            contact=find_contact_iso_trace(pl,four_l_points,further_ept,constrain)
            contact_list.append(contact)
        
        draw_list=[contact_list[0]]
        draw_temp=None
        while len(iso_points)!=0:
            dis_=[]
            for point in iso_points:
                dis=distance_between_pts(point,draw_list[-1])
                dis_.append(dis)
            closest_pt=iso_points[np.argmin(dis_)]

            contact_l=np.round(find_line_eq(contact_list[0],contact_list[1]))
            draw_l=np.round(find_line_eq(closest_pt,draw_list[-1]))
            if tuple(contact_l)==tuple(draw_l) and point_is_in_two_pts(contact_list[1],closest_pt,draw_list[-1]):
                draw_temp=closest_pt
            else:
                if not point_is_in_two_pts(closest_pt,contact_list[0],contact_list[1]):
                    draw_list.append(closest_pt)
            iso_points.remove(closest_pt)
        
        if type(draw_temp)!=type(None):
            draw_list.append(draw_temp)
        
        draw_list.append(contact_list[1])

        
        for i in range(len(draw_list)-1):
            self.msp.add_line(draw_list[i],draw_list[i+1],dxfattribs={'linetype':'DASHDOT', 'layer':draw_layer})

        return contact_list[0],contact_list[1],diag_helper_cut_pt
    def draw_rest_iso(self,iso_size,cut_width):
        for pin in self.pin_at:
            if not pin in self.connected_pins:
                lines=isolation_box_linepts(iso_size,pin)
                for l in lines:
                    self.msp.add_line(l[0],l[1],dxfattribs={'linetype':'DASHDOT', 'layer':'Circuit_Cut'})
                outer_helper=[lines[0][1],lines[-1][-1]]
                
                lines=isolation_box_linepts(iso_size-cut_width,pin)
                for l in lines:
                    self.msp.add_line(l[0],l[1],dxfattribs={'linetype':'DASHDOT', 'layer':'Circuit_Etc'})
                    lines=isolation_box_linepts(iso_size,pin)
                

                lines=isolation_box_linepts(iso_size-(cut_width*2),pin)
                for l in lines:
                    self.msp.add_line(l[0],l[1],dxfattribs={'linetype':'DASHDOT', 'layer':'Circuit_Cut'})
                inner_helper=[lines[0][1],lines[-1][-1]]
                for i in range(len(outer_helper)):
                    self.msp.add_line(outer_helper[i],inner_helper[i],dxfattribs={
                        'layer':'Circuit_Cut',
                        'linetype':'DASHDOT'})
                    
def find_contact_iso_trace(trace_line,iso_l_pts,further_ept,constrain=True):
    inter_temp=[]
    for l_pts in iso_l_pts:
        pt1=l_pts[0]
        pt2=l_pts[1]
        l=find_line_eq(pt1,pt2)
        inter=find_intersect(l,trace_line)
        if type(inter)!=type(None):
            if constrain:
                if point_is_in_two_pts(inter,pt1,pt2):
                    inter_temp.append(inter)
            else:
                inter_temp.append(inter)
    dis_arr=[]
    
    for p in inter_temp:
        dis=distance_between_pts(p,further_ept)
        dis_arr.append(dis)
    contact=inter_temp[np.argmin(dis_arr)]
    return contact

def isolation_box_linepts(box_size,pin_loc,return_line=True):
    """
    box_size= lenght of isoloation box
    """
    tipat=box_size/2
    x=pin_loc[0]
    y=pin_loc[1]
    a=(x-tipat,y-tipat)
    b=(x-tipat,y+tipat)
    c=(x+tipat,y+tipat)
    d=(x+tipat,y-tipat)
    l1=[a,b]
    l2=[b,c]
    l3=[d,c]
    l4=[a,d]
    if return_line:
        return [l1,l2,l3,l4]
    else:
        return [a,b,c,d]
def cross_cut(msp,cross_size,pin_loc):
    """
    cross_size= lenght of bounding box of the cross cut
    """
    tipat=cross_size/2

    x=pin_loc[0]
    y=pin_loc[1]
    cross_s1=(x-tipat,y-tipat)
    cross_e1=(x+tipat,y+tipat)
    cross_s2=(x-tipat,y+tipat)
    cross_e2=(x+tipat,y-tipat)
    msp.add_line(cross_s1,cross_e1,dxfattribs={'layer':'Cut'})       
    msp.add_line(cross_s2,cross_e2,dxfattribs={'layer':'Cut'})
    
def distance_between_pts(pt1,pt2):
    x1,y1=pt1[0],pt1[1]
    x2,y2=pt2[0],pt2[1]
    dis= sqrt((x1-x2)**2+(y1-y2)**2)
    return dis    
def point_is_in_two_pts(pt,pt1,pt2,tolerance=0.05):
   
    x=pt[0]
    y=pt[1]
    x1=np.min([pt1[0],pt2[0]])-tolerance
    x2=np.max([pt1[0],pt2[0]])+tolerance
    y1=np.min([pt1[1],pt2[1]])-tolerance
    y2=np.max([pt1[1],pt2[1]])+tolerance
    if x>= x1 and x<=x2 and y>=y1 and y <=y2:
        return True
    else:
        return False
def find_line_eq(pt1,pt2):
    """
    input: two points on a line
    pt1:[x1,y1]
    pt2:[x2,y2]
    return: a,b,theta in terms of cy=ax+b
    """
    
    x1,y1=pt1[0],pt1[1]
    x2,y2=pt2[0],pt2[1]
    if x1>x2:
        x1,y1=pt2[0],pt2[1]
        x2,y2=pt1[0],pt1[1]

    if x1==x2:
        theta=np.pi/2
        #x= number
        a=1
        b=-x1
        c=0
    else:
        theta=np.arctan((y2-y1)/(x2-x1))
        A=np.array([[x1,1],[x2,1]])
        k=np.array([y1,y2])
        sol=np.linalg.solve(A,k)
        a,b=sol[0],sol[1]
        c=1

    return a,b,c,theta
def find_intersect(line_eq1,line_eq2):
    """
    line_equation follows same pattern as find_line_eq()
    eq=[a,b,c,theta]
    which refers to cy=ax+b
    """
    a1,b1,c1=line_eq1[0],line_eq1[1],line_eq1[2]
    a2,b2,c2=line_eq2[0],line_eq2[1],line_eq2[2]
    if c1==0 and c2==0:
        return None
    
    if a1==a2 and c1!=0 and c2!=0:
        #parallel lines
        return None
    else:
        A=np.array([[a1,-c1],[a2,-c2]])
        k=np.array([-b1,-b2])
        sol=np.linalg.solve(A,k)
        return sol

def find_bi_sec_by_pts(pt1,pt2,pt3):
    """
    return angle bisector by inputing three points (2 lines)
    assume pt2 is the intersect of two interval
    line_equation follows same pattern as find_line_eq()
    eq=[a,b,c,theta]
    """
    x1,x2,x3=pt1[0],pt2[0],pt3[0]
    y1,y2,y3=pt1[1],pt2[1],pt3[1]
    
    if x1==x2 and x2==x3:
        #vertical lines
        bi=[[0,y2,1,0]]
    elif y1==y2 and y2==y3:
        #horizontal lines
        bi=[[1,-x2,0,np.pi/2]]
    else:
        l1=find_line_eq(pt1,pt2)
        l2=find_line_eq(pt2,pt3)
        bi=find_angle_bisector(l1,l2)
    return bi
def find_angle_bisector(line_eq1,line_eq2):
    """
    return an array with 2 angle bisector for 2 straight lines
    line_equation follows same pattern as find_line_eq()
    eq=[a,b,c,theta]
    which refers to cy=ax+b
    """
    theta1=line_eq1[3]
    theta2=line_eq2[3]
    inter_pt=find_intersect(line_eq1,line_eq2)
   
    theta=(theta1+theta2)/2
    theta_=(theta1+theta2)/2 +np.pi/2
    a=np.tan(theta)
    a_=np.tan(theta_)
    b=inter_pt[1]-(a*inter_pt[0])
    b_=inter_pt[1]-(a_*inter_pt[0])
     #to keep format consisent
    c=1

    bi1=[a,b,c,theta]
    bi2=[a_,b_,c,theta_]
    bi=[bi1,bi2]    
    return bi
def find_offset(line_eq,line_width):
    """
    return two lines that offset both out and in with width
    line_equation follows same pattern as find_line_eq()
    eq=[a,b,c,theta]
    """
    halfw=line_width/2
    a,b,c,theta=line_eq[0],line_eq[1],line_eq[2],line_eq[3]
    if c==0:
        scale=1
    else:
        scale=np.cos(theta)
    offset=halfw/scale
    b1=b+offset
    b2=b-offset
    off_l1=[a,b1,c,theta]
    off_l2=[a,b2,c,theta]
    return off_l1,off_l2
def if_two_pt_on_diff_side(line_eq,pt1,pt2):
    """
    return: booleen
    True if two points are on different side of a line
    line_equation follows same pattern as find_line_eq()
    eq=[a,b,c,theta]
    
    """
    a,b,c=line_eq[0],line_eq[1],line_eq[2]
    x1,y1=pt1[0],pt1[1]
    x2,y2=pt2[0],pt2[1]
    if x1>x2:
        x1,y1=pt2[0],pt2[1]
        x2,y2=pt1[0],pt1[1]
    if c!=0:
        _y1=a*x1+b
        _y2=a*x2+b

        if np.sign(_y1-y1)!= np.sign(_y2-y2):
            return True
        else:
            return False
    else:
        x=-b
        if np.sign(x1-x)!=np.sign(x2-x):
            return True
        else:
            return False
        
def if_pts_on_line(line_eq,pt):
    """
    line_equation follows same pattern as find_line_eq()
    eq=[a,b,c,theta]
    return True if point is on the line
    """
    x,y=pt[0],pt[1]
    a,b,c=line_eq[0],line_eq[1],line_eq[2]
    if c==0 and a==1:
        if round(x,2)==round(-b,2):
            return True
        else: return False
    else:
        if round(c*y,2)==round(a*x+b,2):
            return True
        else: return False
def if_pts_are_vert_or_horz(pt1,pt2):
    """
    return True if the line is vertical or horizontal
    """
    [a,b,c,theta]=find_line_eq(pt1,pt2)
    if c==0 or a==0:
        return True
def if_pts_on_interval(pt,pt1,pt2):
    """
    return True if pt is on the inverval pt1 to pt2
    """
    l=find_line_eq(pt1,pt2)
    if if_pts_on_line(l,pt) and point_is_in_two_pts(pt,pt1,pt2):
        return True
    else:
        return False

def dis_between_pt_interval(pt,pt1,pt2):
    l=find_line_eq(pt1,pt2)
    x,y=pt[0],pt[1]
    if if_pts_on_line(l,pt):
        if point_is_in_two_pts(pt,pt1,pt2):
            dis=0
        else:
            dis=np.min([distance_between_pts(pt,pt1),distance_between_pts(pt,pt2)])
    else:
        [a,b,c,theta]=l
        if c!=0:
            if a!=0:
                a_=-a
                c_=1
                b_=y-a_*x
                
            else:
                c_=0
                a_=1
                b_=-x
        else:
            a_=0
            b_=y
            c_=1
        l2=[a_,b_,c_,theta]
        inter=find_intersect(l,l2)
        if if_pts_on_interval(inter,pt1,pt2):
            dis=distance_between_pts(inter,pt)
        else:
            dis=np.min([distance_between_pts(pt,pt1),distance_between_pts(pt,pt2)])
    return dis

def three_ptw_on_one_line(pt1,pt2,pt3):
    l1=find_line_eq(pt1,pt2)
    if if_pts_on_line(l1,pt3):
        return True

def point_within_pinbox(pt,pin,box_size):
    """
    pt=further point
    pin= pin_location
    """
    x,y=pt[0],pt[1]
    size=box_size/2
    xmin=pin[0]-size
    xmax=pin[0]+size
    ymin=pin[1]-size
    ymax=pin[1]+size

    if x>xmin and x<xmax and y>ymin and y<ymax:
        return True
    else: return False

